

set -ex



pytensor-cache help
pip check
python check-for-warnings.py allowed-warnings-base.txt
python -m pytest tests/test_breakpoint.py tests/test_config.py tests/test_gradient.py tests/test_ifelse.py tests/test_raise_op.py tests/test_rop.py tests/test_updates.py
exit 0
